"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
from ..exceptions import PinkyError

# XXX FIX ME
#   this would be much better moved into a class structure
#   a lot of data structures are used between the functions
NEVER = -1
MAYBE = 0
AROMATIC = 1
AROMATIC_ATOMS = {'N':1, 'C':1, 'S':1, 'O':1}
AROMATIC_PYROLE_ATOMS = {'C':1, 'N':1, 'O':1, 'S':1}
AROMATIC_5_RING = [(1,), (2,4), (1,4), (2,4), (1,)]

# Here's the deal.  When we get molecules out of the smiles
# parser, some properties are not known.  Mainly, we would
# like to find whether
#  1) is the bond type aromatic or single for non specified
#     bonds.  Non specified bonds are like C1CCCCC1 in the
#     smiles strings.  These are determined through the bond.fixed
#     property.  If a bond is fixed then it really is a single, double
#     triple or sacrifice fly bond :)
#  2) is an atom aromatic or not?
#
# At this point what is known is this:
#  lower case atoms are specified to be aromatic.
#  These have atom.aromatic = 1
#
#  Bonds are single, double or triple.
#   a double or triple bound can never be changed.
#   a specified single bond can never be changed
#   

# Table that indicates which atoms can be considered
# "pyrole like" and behave like the pyrole nitrogen
# in c1[nH]ccc1
#symbol charge hcount bonds
PyroleTable = {
    ('C', -1, 1, 2):1,
    ('C', -1, 0, 3):1,
    ('N', 0, 1, 2):1,
    ('N', 0, 0, 3):1,
    ('O', 0, 0, 2):1,
    ('S', 0, 0, 2):1,
    }

def getPyroleLikeAtoms(cycle):
    """cycle->return a dictionary of pyrole nitrogen-like atoms in
    a cycle or a molecule  The dictionary is keyed on the atom.handle"""    
    result = {}
    # the outgoing bonds might need to be single or aromatic
    for atom in cycle.atoms:
        lookup = (atom.symbol, atom.charge, atom.hcount, len(atom.bonds))
        if PyroleTable.get(lookup, 0):
            result[atom.handle] = atom

    return result
        
def canBeAromatic(cycle, pyroleLike):
    """(cycle)-> returns AROMATIC if a ring is conjugatable and
                                  passes the simple tests for aromaticity
                 returns MAYBE if the ring in its present form
                                can be aromatic but is not currently
                         NEVER if the ring can never be aromatic"""
    cycleLength = len(cycle)
    # *******************************************************
    #  check for kekular five membered rings
    if cycleLength == 5:
        # check atom types
        for atom in cycle.atoms:
            if not atom.symbol in AROMATIC_PYROLE_ATOMS:
                return NEVER

        # do we have exactly one pyrole nitrogen like atom?
        pyroleCount = 0
        for atom in cycle.atoms:
            if atom.handle in pyroleLike:
                pyrole = atom
                pyroleCount += 1

        
        if pyroleCount < 1 or pyroleCount > 2:
            return NEVER

        # rotate the ring so that we start on the pyrole like atom
        cycle.rotate(pyrole)
        bonds = cycle.bonds[:]
        # check the bonds for a kekular structure
        for index, bond in zip(range(len(bonds)), bonds):
            if bond.bondtype not in AROMATIC_5_RING[index]:
                return MAYBE
            
        return AROMATIC
    # *****************************************************
    #  check for kekular six membered rings
    #  kekular rings must have atoms in the AROMATIC_ATOMS
    #  groups and must belong in 6 membered rings.
    #  bonds must be conjugated
    elif cycleLength == 6:
        # XXX FIX ME -> there is a lot of problems with this
        # code I think, what about bonds that are already fixed?
        for atom in cycle.atoms:
            if not atom.symbol in AROMATIC_ATOMS:
                return NEVER
            
        bonds = cycle.bonds[:]        
                
        last = None
        switch = {1:2, 2:1}
        while bonds:
            bond = bonds.pop()
            bondtype = bond.bondtype

            if bond.bondorder == 3:
                return NEVER
            
            if last is None:
                if bond.bondtype in [1,2]:
                    last = bond.bondtype
            else:
                if last == 1 and bond.bondtype not in [2,4]:
                    return MAYBE
                elif last == 2 and bond.bondtype not in [1, 4]:
                    return MAYBE

                last = switch[last]
                if bondtype != last:
                    bond.bondorder = last

        return AROMATIC
    
    else:
        # we can never be aromatic
        return NEVER

def convert(cycle, pyroleLike, usedPyroles):
    """cycle, pyroleLike, aromatic=0-> aromatize the cycle
    pyroleLike is a lookup of the pyrole like atoms in the
    cycle.
    return 1 if the cycle was aromatized
           2 if the cycle could not be aromatized"""

    bonds = cycle.bonds
    atoms = cycle.atoms
    initialBondStates = []
    initialAtomStates = []
    _usedPyroles = {}
    for bond in bonds:
        # store the initial states but assume the
        # bond is aromatic
        initialBondStates.append((bond, bond.symbol,
                                  bond.bondorder, bond.bondtype,
                                  bond.aromatic, bond.stereo))
        # XXX FIX ME
        # until we get proper conjugation, aromatic bond orders
        # are 1.5
        bond.reset(':', bond.bondorder, 4, bond.fixed, bond.stereo)
        
    aromatized = 1
    for atom in atoms:
        initialAtomStates.append((atom, atom.aromatic))
        atom.aromatic = 1

        nonhydrogens = atom.sumBondOrders() + atom.charge

        # look for the lowest valence where we don't
        #  have to change the charge of the atom to
        #  fill the valences
        for valence in atom.valences:
            neededHydrogens = int(valence - nonhydrogens)
            if neededHydrogens >= 0:
                break
        else:
            # we can't change the aromaticity and have correct
            # valence.
            #
            # there is one special case of a five membered
            #  ring and a pyrole nitrogen like atom we need
            #  to look for.
            if len(cycle) == 5 and atom.handle in pyroleLike:
                _usedPyroles[atom.handle] = 1
            else:
                # nope, the valences don't work out so
                # we can't aromatize
                aromatized = 0
                break

    # sanity check, this should be true because of the
    # canBeAromatic routine above
    assert len(_usedPyroles) <=1, "Too many used pyroles!"
    
    cycle.aromatic = aromatized
    if not aromatized:
        for bond, symbol, order, bondtype, aromatic, stereo in initialBondStates:
            bond.reset(symbol, order, bondtype, bond.fixed, stereo)

        for atom, aromatic in initialAtomStates:
            atom.aromatic = aromatic
    else:
        # we used some pyroles, we'll have to send these to
        # the valence checker later
        usedPyroles.update(_usedPyroles)

    return aromatized

def addHydrogens(molecule, usedPyroles=None):
    """(molecule) -> add implicit hydrogens to a molecule.
    If the atom has specified valences and the atom must be
    charged then a Valence Error is raised"""
    for atom in molecule.atoms:
        # if the atom has an explicit hcount, we can't set the
        # hcount
        if atom.has_explicit_hcount:
            atom.hcount = atom.explicit_hcount
            continue
        
        if atom.valences:            
            for valence in atom.valences:
                hcount = max(0, int(valence - atom.sumBondOrders() + atom.charge))
                if hcount >= 0:
                    break
            else:
                if usedPyroles and not atom.handle in usedPyroles:
                    #print atom.symbol, atom.valences, atom.hcount, atom.charge,\
                    #      atom.sumBondOrders()
                    #print [x.bondtype for x in atom.bonds]
                    #print molecule.cansmiles()
                    raise PinkyError("Valence error in atom %s"%molecule.atoms.index(atom))
                pass

            #hcount = int(hcount)
            atom.hcount = hcount
    return molecule

def fixBonds(molecule, usedPyroles):
    bondsToBeFixed = {}

    # collect the bonds that need to be fixed
    for bond in molecule.bonds:
        if bond.fixed == 0:
            for atom in bond.atoms:
                if not atom.rings:
                    bond.fixed = 1
                    break
            else:
                bondsToBeFixed[bond.handle] = bond

    if not bondsToBeFixed:
        return molecule
    
    # seperate the bonds in cycles from the bonds outside
    # XXX FIX ME, this is a slow way to do this, perhaps
    # the bond should "know" what cycles it's in?
    # cycleBonds holds bonds in aromatic cycles
    cycleBonds = {}
    for cycle in molecule.cycles:
        if not cycle.aromatic:
            # if the cycle is not aromatic, assume
            # that the bond orders are already okay
            for bond in cycle.bonds:
                bond.fixed = 1
                if bond.handle in bondsToBeFixed:
                    del bondsToBeFixed[bond.handle]
        else:
            for bond in cycle.bonds:
                # check for bonds adjacent to pyrole like
                # atoms
                if bond.handle in bondsToBeFixed and len(cycle) == 5:
                    for atom in bond.atoms:
                        if atom in cycle.atoms and atom.handle in usedPyroles:
                            bond.reset(bond.symbol, 1, 4, 1, bond.stereo)
                            del bondsToBeFixed[bond.handle]
                            break
                cycleBonds[bond.handle] = bond

    for bond in bondsToBeFixed.values():
        if not bond.handle in cycleBonds:
            if bond.bondtype == 4:
                raise PinkyError("Aromatic Bond outside ring %s"% bond)
            else:
                # fix the bond
                # precondition, bondorder and bondtype are the
                # same!
                assert bond.bondorder == bond.bondtype
                bond.reset(bond.symbol, bond.bondorder,
                           bond.bondtype, 1, bond.stereo)
##  This below seems wrong, it doesn't work!
##        else:
##            for atom in bond.atoms:
##                # XXX FIX ME, should I check for all valences?
##                assert atom.valences
##                if atom.sumBondOrders() + atom.explicit_hcount >= atom.valences[0]:
##                    assert not bond.fixed
##                    bond.bondorder = 1
##                    assert bond.bondtype == 4 or bond.bondtype == bond.bondorder, "bondtype = %s, bondorder=%s"%(
##                        bond.bondtype, bond.bondorder)

##                    print "setting valence", bond
##                    bond.fixed = 1
##                    del bondsToBeFixed[bond.handle]
##                    break
                    
    while bondsToBeFixed:
        changed = 1
        while changed:
            changed = 0

            for bond in list(bondsToBeFixed.values()):
                for atom in bond.atoms:
                    for connectedBond in atom.bonds:
                        if connectedBond is not bond and \
                               connectedBond.handle in cycleBonds and \
                               connectedBond.fixed:
                            if connectedBond.bondorder == 2:
                                bondorder = 1
                            else:
                                bondorder = 2

                            bond.bondorder = bondorder
                            bond.fixed = 1
                            changed = 1
                            del bondsToBeFixed[bond.handle]
                            break
                    # if we fixed the bond, break out of the atom
                    # loop
                    if bond.fixed: break

        if not changed and bondsToBeFixed:
            # arbitrarily fix a bond to be a single bond
            bond = list(bondsToBeFixed.values())[0]
            bond.fixed = 1
            del bondsToBeFixed[bond.handle]
                        
    for bond in molecule.bonds:
        assert bond.fixed == 1
    return molecule
        
def aromatize(molecule, usedPyroles=None):    
    """(molecule, usedPyroles=None)->aromatize a molecular graph
    usedPyroles is a dictionary that holds the pyrole like
    atoms that are used in the conversion process.
    The following valence checker may need this information"""
    
    pyroleLike = getPyroleLikeAtoms(molecule)

    if usedPyroles is None:
        usedPyroles = {}
        
    cyclesToCheck = []
    # determine which cycles came in marked as aromatic
    # and which need to be checked form the kekular form
    #
    # if a cycle came in as aromatic, convert it
    # before going on.
    for cycle in molecule.cycles:
        for atom in cycle.atoms:
            if not atom.aromatic:
                cyclesToCheck.append(cycle)
                break
        else:
            if not convert(cycle, pyroleLike, usedPyroles):
                # XXX FIX ME
                # oops, an aromatic ring came in but
                # we can't convert it.  This is an error
                # daylight would conjugate the ring
                raise PinkyError("Bad initial aromaticity")

    # keep checking rings until something happens
    while 1:
        # assume nothing happened
        needToCheckAgain = 0

        _cyclesToCheck = []
        for cycle in cyclesToCheck:
            canAromatic = canBeAromatic(cycle, pyroleLike)
            if canAromatic == NEVER:
                # the ring can NEVER EVER be aromatic, so remove it for good
                pass
            elif canAromatic and convert(cycle, pyroleLike, usedPyroles):
                needToCheckAgain = 1
            else:
                _cyclesToCheck.append(cycle)

        cyclesToCheck = _cyclesToCheck
        if not needToCheckAgain:
            break

    # fix bonds that have no bondorder if necessary
    molecule = fixBonds(molecule, pyroleLike)
    # add implicit hydrogens
    return addHydrogens(molecule, usedPyroles)



from unittest import TestCase
from nose.tools import *
from pinky.smiles import smilin
from pinky.fingerprints import ecfp

class ECFPTestCase(TestCase):
    def test_ecfp(self):
        """Test Butyramide (example used in ECFP paper)"""
        mol = smilin('CCCC(=O)N')

        # ECFP_0
        fp = ecfp(mol, radius=0)
        assert len(fp) == 5

        # ECFP_2
        fp = ecfp(mol, radius=1)
        assert len(fp) == 11

        # ECFP_4
        fp = ecfp(mol, radius=2)
        assert len(fp) == 14

        # ECFP_6
        fp = ecfp(mol, radius=3)
        assert len(fp) == 14



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

# sheesh, how much indirection can there be?
# I'm longing for Mitch's Graph.T right about
# now :)
from ..canonicalization import disambiguate, traverse
from ..canonicalization.equiv import compute_equiv_class


# for now these molecules will be immutable
# Fix Me, don't use _atoms use atom.parent
class Molecule(object):
    def __init__(self, atoms=None, bonds=None):
        self.atoms = atoms or []
        self.bonds = bonds or []
        index = 0
        for atom in self.atoms:
            if atom.parent:
                ValueError("Atom %s already has a parent"%atom)
            atom.parent = self
            atom.index = index
            index += 1
            if atom.chiral_class:
                atom.setchival(None, atom.chiral_class)
        index = 0
        for bond in self.bonds:
            if bond.parent:
                ValueError("Atom %s already has a parent"%atom)                
            bond.parent = self
            bond.index = index
            index += 1
            
        self.fields = {}
        self.cycles = []

        self._canonical = None
        self.dirty = 1   # a molecule is dirty
                         #  when it needs to be recanonicalized
                         #  or various properties need to be updated

        self.name = ""
        self.vfgraph = None  # place holder for a vflib matching graph
                             # must be rebuilt if None
                            
    def add_atom(self, atom):
        if atom.parent:
            raise("Already have atom", atom)

        # set the parent of the atom to be this object
        # set up all the relevant datastructures for
        # the atom
        atom.parent = self
        atom.index = len(self.atoms)
        
        self.atoms.append(atom)
        self.dirty = 1
        if self.vfgraph:
            index = self.vfgraph.InsertNode(atom)
            assert index == atom.index
        self.vfgraph = None

    def add_bond(self, bond, atom1, atom2):
        if atom1.parent is not self:
            raise("Atom", atom1, "is not in the molecule")
        if atom2.parent is not self:
              raise("Atom", atom2, "is not in the molecule")          
        if bond.parent:
            raise("Bond", bond, "already has a parent")
        
        bond.parent = self
        bond.index = len(self.bonds)        
        self.bonds.append(bond)
        bond.atoms = [atom1, atom2]

        atom1.bonds.append(bond)
        atom2.bonds.append(bond)
        atom1.oatoms.append(atom2)
        atom2.oatoms.append(atom1)

        self.dirty = 1
        if self.vfgraph:
            vfgraph = self.vfgraph
            vfgraph.InsertEdge(index1, index2, bond)
            vfgraph.InsertEdge(index2, index1, bond)

    def remove_bond(self, bond):
        self.bonds.remove(bond)
        bond.destroy()
        self.dirty = 1
        self.vfgraph = None
        index = 0
        for bond in self.bonds:
            bond.index = index
            index += 1

    def remove_atom(self, atom):
        if atom.parent is not self:
            raise "Atom not in molecule"

        for bond in atom.bonds[:]:
            self.remove_bond(bond)

        self.atoms.remove(atom)        
        atom.parent = None
        atom.destroy()
        self.dirty = 1
        self.vfgraph = None
        index = 0
        for atom in self.atoms:
            atom.index = index
            index += 1
            
    def cansmiles(self, isomeric=0):
        if isomeric: draw = traverse.drawIsomeric
        else: draw = traverse.draw
        
        if self.dirty:
            # XXX FIX ME
            # Move equivalence_class setter outside?
            # recompute the equivalence classes
            for atom in self.atoms:
                atom.equiv_class = compute_equiv_class(atom)
            disambiguate.FreedDisambiguate(self)
            self._canonical, self.canonical_list = traverse.draw(self)
            self.dirty = 0

        if isomeric:
            return traverse.drawIsomeric(self)[0]
        return self._canonical

    def arbsmarts(self, isomeric=0):
        # XXX FIX ME
        # compute canonicalization smarts
        # without hydrogens and charges
        canonical, canonical_list = traverse.drawSmarts(self)
        return canonical
            
    def arbsmiles(self, isomeric=0):
        if isomeric: draw = traverse.drawIsomeric
        else: draw = traverse.draw
        
        symorders = [atom.symorder for atom in self.atoms]

        # XXX FIX ME - kind of a hack!
        # set all the symorders to the index of the
        #  atom in self.atoms
        i = 0
        for atom in self.atoms:
            atom.symorder = i
            i+= 1
            
        result, self.arb_list = draw(self)

        # restore the symorders
        for atom, order in zip(self.atoms, symorders):
            atom.symorder = order

        return result    

    def pruneToAtoms(self, atoms):
        """Prune the molecule to the specified atoms
        bonds will be removed atomatically"""
        _atoms = self.atoms[:]
        for atom in _atoms:
            if atom not in atoms:
                self.remove_atom(atom)



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



Traversals - a traversal (or path) through a graph

Traversals are stored as a list of tokens representing each object in a graph.
Each token object has __str__ defined so once a list has been formed
it can be turned into a string object by

 "".join(traversal_list)

Tokens are defined by the representation desired.  See tokens.py for
a list of SmilesTokens.

One thing to remember is that closures are considered atom properties for
the purposes of traversals.  That is the AtomToken is reponsible for storing
information about closures.  This information is built up during the
depth first search of the graph.
"""

from pinky.canonicalization import tokens

class IDGenerator:
    def __init__(self):
        self.index = 0

    def next(self):
        self.index += 1
        return self.index
    
class SmilesTraversal:
    # override the following to create different
    # tokens for traversals
    AtomToken = tokens.Atom
    BondToken = tokens.Bond
    BranchToken = tokens.Branch
    BranchEndToken = tokens.BranchEnd

    def __init__(self, parent=None):
        if parent is None:
            atomsDone = {}
            idGenerator = IDGenerator()
            closureIdGenerator = IDGenerator()
            closures = {}
        else:
            atomsDone = parent.atomsDone
            idGenerator = parent.idGenerator
            closureIdGenerator = parent.closureIdGenerator
            closures = parent.closures

        self.atoms = []
        self.bonds = []
        self.data = []
        self.atomsDone = atomsDone
        self.idGenerator = idGenerator
        self.closureIdGenerator = closureIdGenerator
        self.closures = closures

    def addAtom(self, atom):
        #print "atom", atom,
        atomToken = self.AtomToken(atom,
                                   self.closures,
                                   self.closureIdGenerator)
        self.atomsDone[atom] = atomToken
        self.data.append(atomToken)
        self.atoms.append(atom)
        #print self.atomsDone

    def addBond(self, bond):
        #print "bond", bond
        self.data.append(self.BondToken(bond))
        # XXX HACK -> give the bond a traversal order
        self.bonds.append(bond)

    def addClosure(self, atom1, atom2, bond):
        #print "closure", atom1, atom2, bond
        atomsDone = self.atomsDone
        closures1 = atomsDone[atom1].closures
        closures2 = atomsDone[atom2].closures
        assert closures1 is not closures2
        id = self.idGenerator.next()
        
        bondToken = self.BondToken(bond)
        # only pass the bond info to the first atom
        # the other gets a None (or don't print bond info)
        closures1.append((id, bondToken))
        closures2.append((id, None))
        self.bonds.append(bond)

    def addBranch(self):
        self.data.append(self.BranchToken())

    def addBranchEnd(self):
        self.data.append(self.BranchEndToken())

    def append(self, traverse):
        """(traverse)->append the traverse to the current traverse"""
        self.data.extend(traverse.data)
        self.atoms.extend(traverse.atoms)
        self.bonds.extend(traverse.bonds)

    def __str__(self):
        # XXX HACK -> find the bond traversal order
        # and make a note of it.  This is pretty ugly
        index = 1
        for x in self.data:
            if x.__class__ == tokens.Bond:
                x.bond._traverseOrder = index
                index += 1
        return "".join(map(str,self.data))


class SmartsTraversal(SmilesTraversal):
    AtomToken = tokens.SmartsAtom
    BondToken = tokens.SmartsBond

class IsomericSmilesTraversal(SmilesTraversal):
    AtomToken = tokens.IsomericAtom



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import re
import string
from pinky.smiles import handler
from .builder import BuildMol
from ..perception import aromaticity, figueras

# To verify this is correct, run
#  support.make_re_pattern(support.element_symbols + support.aromatic_symbols)
# Yes, I included aromatics in this list.
element_symbols_pattern = \
  r"C[laroudsemf]?|Os?|N[eaibdpos]?|S[icernbmg]?|P[drmtboau]?|"  \
  r"H[eofgas]?|c|n|o|s|p|A[lrsgutcm]|B[eraik]?|Dy|E[urs]|F[erm]?|"  \
  r"G[aed]|I[nr]?|Kr?|L[iaur]|M[gnodt]|R[buhenaf]|T[icebmalh]|" \
  r"U|V|W|Xe|Yb?|Z[nr]|\*"

atom_fields = [
    "raw_atom",
    "open_bracket",
    "weight",
    "element",
    "chiral_count",
    "chiral_named",
    "chiral_symbols",
    "hcount",
    "positive_count",
    "positive_symbols",
    "negative_count",
    "negative_symbols",
    "error_1",
    "error_2",
    "close_bracket",
    "error_3",
    ]

atom = re.compile(r"""
(?P<raw_atom>Cl|Br|[cnospBCNOFPSI]) |  # "raw" means outside of brackets
(
  (?P<open_bracket>\[)                 # Start bracket
  (?P<weight>\d+)?                     # Atomic weight (optional)
  (                                    # valid term or error
   (                                   #   valid term
    (?P<element>""" + element_symbols_pattern + r""")  # element or aromatic
    (                                  # Chirality can be
     (?P<chiral_count>@\d+) |          #   @1 @2 @3 ...
     (?P<chiral_named>                 # or
       @TH[12] |                       #   @TA1 @TA2
       @AL[12] |                       #   @AL1 @AL2
       @SP[123] |                      #   @SP1 @SP2 @SP3
       @TB(1[0-9]?|20?|[3-9]) |        #   @TB{1-20}
       @OH(1[0-9]?|2[0-9]?|30?|[4-9])) | # @OH{1-30}
     (?P<chiral_symbols>@+)            # or @@@@@@@...
    )?                                 # and chirality is optional
    (?P<hcount>H\d*)?                  # Optional hydrogen count
    (                                  # Charges can be
     (?P<positive_count>\+\d+) |       #   +<number>
     (?P<positive_symbols>\++) |       #   +++...  This includes the single '+'
     (?P<negative_count>-\d+)  |       #   -<number>
     (?P<negative_symbols>-+)          #   ---...  including a single '-'
    )?                                 # and are optional
    (?P<error_1>[^\]]+)?               # If there's anything left, it's an error
  ) | (                                # End of parsing stuff in []s, except
    (?P<error_2>[^\]]*)                # If there was an error, we get here
  ))
  ((?P<close_bracket>\])|              # End bracket
   (?P<error_3>$))                     # unexpectedly reached end of string
)
""", re.X)

bond_fields = ["bond"]
bond = re.compile(r"(?P<bond>[=#/\\:~-])")

dot_fields = ["dot"]
dot = re.compile(r"(?P<dot>\.)")

closure_fields = ["closure"]
closure = re.compile(r"(?P<closure>\d|%\d\d?)")

close_branch_fields = ["close_branch"]
close_branch = re.compile(r"(?P<close_branch>\))")

open_branch_fields = ["open_branch"]
open_branch = re.compile(r"(?P<open_branch>\()")

# Make from the state name to
#   1. the regular expession to try to match and
#   2. the list of field names in that regexp.
# (There's no way to use sre to get #2 as an ordered list so
#  it needs to be done manually.)
info = {
    "atom": (atom, atom_fields),
    "bond": (bond, bond_fields),
    "dot": (dot, dot_fields),
    "closure": (closure, closure_fields),
    "close_branch": (close_branch, close_branch_fields),
    "open_branch": (open_branch, open_branch_fields),
    }

# Mapping from current state to allowed states
table = {
    # Could allow a dot
    "start": ("atom",),

    # CC, C=C, C(C)C, C(C)C, C.C, C1CCC1
    "atom": ("atom", "bond", "close_branch", "open_branch", "dot", "closure"),

    # C=C, C=1CCC=1
    "bond": ("atom", "closure"),

    # C(C)C, C(C)=C, C(C).C, C(C(C))C, C(C)(C)C
    "close_branch": ("atom", "bond", "dot", "close_branch", "open_branch"),

    # C(C), C(=C), C(.C) (really!)
    "open_branch": ("atom", "bond", "dot"),

    # C.C  -- allow a dot? as in C..C
    "dot": ("atom",),

    # C1CCC1, C1=CCC1, C12CC1C2, C1C(CC)1, C1(CC)CC1, c1ccccc1.[NH4+]
    "closure": ("atom", "bond", "closure", "close_branch", "open_branch", "dot"),
}

# Parse a SMILES string and print the events found
def tokenize(s, handler = handler.TokenHandler()):
    expected = table["start"]
    n = len(s)
    i = 0
    handler.begin()
    while i < n:
        # Of the expected states, find one that matches the
        # text at the current position.
        for state in expected:
            pat, fields = info[state]
            m = pat.match(s, i)
            if m:
                break
        else:
            # No matches found, so this was an error
            handler.error("Unknown character", i, s[i:])
            # The handler is allowed to not throw an
            # exception, but we are done, so return.
            return
        #print "New state:", state,

        # Get the dictionary of matched name groups
        d = m.groupdict()

        # Go through the list of fields that could have matched.
        # Needs to go in a order so the token text can be converted
        # back into the original string.
        for field in fields:
            # See if there was a match for the given named field
            if d[field] is not None:
                # Was it an error match?
                if field[:5] == "error":
                    pos = m.start(field)
                    if field == "error_3":
                        handler.error("Missing ']'", pos, s[pos:])
                    else:
                        handler.error("Unknown character", pos, s[pos:])
                    return
                # Success, so send the token to the callback
                #print "--> ", m.group(field), field
                handler.add_token(field, i, m.group(field))

        # Get the new list of expected states, and move to
        # the end of the previous match.
        expected = table[state]
        i = m.end(0)

    handler.end()

def smilin(smiles, transforms=[figueras.sssr, aromaticity.aromatize]):
    """(smiles)->molecule
    Convert a smiles string into a molecule representation"""
    builder = BuildMol()
    tokenize(smiles, builder)
    mol = builder.mol

        
    for transform in transforms:
        mol = transform(mol)

    ## implicit hcount doesn't make any sense anymore...
    for atom in mol.atoms:
        if not atom.has_explicit_hcount:
            atom.imp_hcount = atom.hcount - atom.explicit_hcount

    return mol



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Disambiguate

FreedDisambiguate(graph)

given atom and bond equiv_classes use Freed's technique to
find symmettry classes and symmetry orders for the graph

These were initial described in the following papers published in
JCICS
   1. WEININGER D,
     "SMILES, A chemical language and information-system. 1.
      Introductions to methodology and encoding rules"

      JOURNAL OF CHEMICAL INFORMATION AND COMPUTER SCIENCES,28(#1),
      1988,31-36 
   2. WEININGER D, WEININGER A, WEININGER JL,
     "SMILES 2. Algorithm for generation of unique SMILES notation"

     JOURNAL OF CHEMICAL INFORMATION AND COMPUTER SCIENCES,29(#2),
     1989,97-101 

They do not talk about canonicalizing using bond types or stereo
chemistry though.  See notes below for details.
"""
from pinky.canonicalization import primes
from itertools import zip_longest

class FreedDisambiguate:
    def __init__(self, graph):
        """(graph)
        
        given atom and bond equiv_classes use Freed's technique to
        find symmettry classes and symmetry orders for the graph
        """
        self.graph = graph
        atoms = graph.atoms
        self.range = range(len(atoms))
        offsets = self.offsets = []

        # symmetry classes are the symmetrical points in a graph
        #  so if two vertices share the same value, they are symmetric
        symclasses = self.symclasses = []

        # symmetry orders describe the traversal order to canonicalize
        # the graph.  All of these should be unique values
        symorders = self.symorders = [0] * len(atoms)
        indices = {}

        # cache the atom indices in a lookup table
        index = 0
        for atom in atoms:
            indices[atom.handle] = index
            index += 1

        # initialize the necessary data
        for atom in atoms:
            # For each atom, get the neighbors offsets and the connected
            # bondtype for a given atom
            # N.B. the oatom (adjacent atom)
            # list and the bond list must be kept in sync
            # bondtype must be an integer!!!!
            
            # the weininger algorithm is modified by also
            # multiplying by not only the neighboring symorder
            #  but by the bondtype used to reach the neighbor.
            #  The traversal algorithm will then traverse to
            #  the lowest symorder atom with the lowest bondtype
            #  first.
            neighbors = zip([indices[x.handle] for x in atom.oatoms],
                            [x.bondtype for x in atom.bonds])

            # store the symorder index of the neighboring atoms
            # and the bondtype used to traverse to each neighbor
            
            # This forms a list of the form
            # (neighbor_symorder_index, bondtype),
            #   ...(neighbor_symorder_index, bondtype) ...
            offsets.append(neighbors)

            # the symmetry classes start out as the atom's equivalence
            # class
            symclasses.append(atom.equiv_class)

        #########################################################
        # find the symmetry classes
        # rank the current symmetry orders 
        self.rank()

        # Find the sym classes invariant for the current
        self.symclasses = self.findInvariant(self.symclasses)

        # now find the symmetry orders from the symmetry classes
        self.symorders = self.symclasses[:]
        
        self.findInvariantPartitioning()

        # Give them back to the atoms and bonds
        symclasses = self.symclasses
        symorders = self.symorders
        i = 0
        for atom in atoms:
            atom.symclass = symclasses[i]
            atom.symorder = symorders[i]
            i = i + 1

    def disambiguate(self, symclasses):
        """Use the connection to the atoms around a given vertex
        as a multiplication function to disambiguate a vertex"""
        offsets = self.offsets
        symclasses = list(symclasses)
        result = symclasses[:]
        for index in self.range:
            try:
                val = 1
                for offset, bondtype in offsets[index]:
                    val *= symclasses[offset] * bondtype
            except OverflowError:
                # Hmm, how often does this occur?
                val = long(1)
                for offset, bondtype in offsets[index]:
                    val *= symclasses[offset] * bondtype

            result[index] = val
        return result

    def rank(self):
        """convert a list of integers so that the lowest integer
        is 0, the next lowest is 1 ...
        note: modifies list in place"""
        # XXX FIX ME, should the lowest value be 1 or 0?
        symclasses = self.symclasses
        stableSort = sorted(zip_longest(symclasses, range(len(symclasses))))

        last = None
        x = -1
        for order, i in stableSort:
            if order != last:
                x += 1
                last = order
            symclasses[i] = x

    def breakRankTies(self, oldsym, newsym):
        """break Ties to form a new list with the same integer ordering
        from high to low

        Example
        old = [ 4, 2, 4, 7, 8]  (Two ties, 4 and 4)
        new = [60, 2 61,90,99]
        res = [ 4, 0, 3, 1, 2]
                *     *        This tie is broken in this case
        """
        stableSort = sorted(zip_longest(oldsym, newsym, range(len(oldsym))))

        lastOld = None
        lastNew = None
        x = -1
        for old, new, index in stableSort:
            if old != lastOld:
                x += 1
                # the last old value was changed, so update both
                lastOld = old
                lastNew = new
            elif new != lastNew:
                # break the tie based on the new info (update lastNew)
                x += 1
                lastNew = new
            newsym[index] = x
    
    def findLowest(self, symorders):
        """Find the position of the first lowest tie in a
        symorder or -1 if there are no ties"""
        _range = range(len(symorders))
        # XXX FIX ME
        # Do I need to sort?
        stableSymorders = sorted(zip_longest(symorders, _range))

        lowest = None
        for index in _range:
            if stableSymorders[index][0] == lowest:
                return stableSymorders[index-1][1]
            lowest = stableSymorders[index][0]
        return -1

    def findInvariant(self, symclasses):
        """(symclasses) -> converge the disambiguity function
        until we have an invariant"""
        get = primes.primes.get
        disambiguate = self.disambiguate
        breakRankTies = self.breakRankTies

        while 1:
            newSyms = map(get, symclasses)
            newSyms = disambiguate(newSyms)
            breakRankTies(symclasses, newSyms)
            if symclasses == newSyms:
                return newSyms
            symclasses = newSyms

    def findInvariantPartitioning(self):
        """Keep the initial ordering of the symmetry orders
        but make all values unique.  For example, if there are
        two symmetry orders equal to 0, convert them to 0 and 1
        and add 1 to the remaining orders

          [0, 1, 0, 1]
        should become
          [0, 2, 1, 3]"""
        
        symorders = self.symorders[:]
        _range = range(len(symorders))
        while 1:
            pos = self.findLowest(symorders)
            if pos == -1:
                self.symorders = symorders
                return
            for i in _range:
                symorders[i] = symorders[i] * 2 + 1
            symorders[pos] = symorders[pos] - 1
            
            symorders = self.findInvariant(symorders)





"""
Implementation of Extended-Connectivity Fingerprints algorithm (David Rogers and Mathew Hahn 2010). 

This is based on the MorganFingerprints.cpp code from rdkit:
  https://github.com/rdkit/rdkit/blob/master/Code/GraphMol/Fingerprints/MorganFingerprints.cpp
"""
from bitarray import bitarray
import copy

def gen_hash(lst):
    return hash(tuple(lst))

def invariants(mol):
    """Generate initial atom identifiers using atomic invariants"""
    atom_ids = {}
    for a in mol.atoms:
        components = []
        components.append(a.number)
        components.append(len(a.oatoms))
        components.append(a.hcount)
        components.append(a.charge)
        components.append(a.mass)
        if len(a.rings) > 0:
            components.append(1)

        atom_ids[a.index] = gen_hash(components)

    return atom_ids

def ecfp(mol, radius=2):
    """Compute the Extended-Connectivity fingerprint for a molecule.
    
    :param mol: molecule object parsed from SMILES string
    :param radius: The number of iterations to perform. Defaults to 2 which is equivilent to ECFP4.
    :rtype: dictionary representing the molecular fingprint (atom identifiers and their counts).
    """

    atom_ids = invariants(mol)

    fp = {}
    for i in atom_ids.values():
        fp[i] = fp.get(i, 0) + 1

    neighborhoods = []
    atom_neighborhoods = [ len(mol.bonds) * bitarray('0') for a in mol.atoms]
    dead_atoms = len(mol.atoms) * bitarray('0')

    for layer in range(1, radius+1):
        round_ids = {}
        round_atom_neighborhoods = copy.deepcopy(atom_neighborhoods)
        neighborhoods_this_round = []

        for a in mol.atoms:
            if dead_atoms[a.index]: continue

            nbsr = []
            for b in a.bonds:
                round_atom_neighborhoods[a.index][b.index] = True
                oidx = b.xatom(a).index
                round_atom_neighborhoods[a.index] |= atom_neighborhoods[oidx]
                nbsr.append((b.bondtype, atom_ids[oidx]))

            nbsr = sorted(nbsr)
            nbsr = [item for sublist in nbsr for item in sublist]
            nbsr.insert(0, atom_ids[a.index])
            nbsr.insert(0, layer)

            round_ids[a.index] = gen_hash(nbsr)
            neighborhoods_this_round.append(
                (round_atom_neighborhoods[a.index], round_ids[a.index], a.index)
            )

        for lst in neighborhoods_this_round:
            if lst[0] not in neighborhoods:
                fp[lst[1]] = fp.get(lst[1], 0) + 1
                neighborhoods.append(lst[0])
            else:
                dead_atoms[lst[2]] = True

        atom_ids = round_ids
        atom_neighborhoods = copy.deepcopy(round_atom_neighborhoods)
    return fp



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Recursively traverse a molecule building up a canonical
representation.

Each atom of a Molecule or Graph must have a attribute 'symorder' which
is a unique number.  This number guarantees only one traversal for
the graph.

Additionally each bond must have an attribute equiv_class which is
a unique value for each different type of bond.  This guarantees
proper canonicalization of bonds as well as atoms.

canonical_string = Traverse.draw(molecule)

Advanced usage:
canonical_string = Traverse.draw(molecule, TraversalType)

TraversalType controls how the traversal is represented.
SmilesTraversal is the default TraversalType but this can be
subclassed for different representations.  For example it is
easy to create a subclass to generate Tripos Line formats.
"""

from .smiles import SmilesTraversal, SmartsTraversal,  IsomericSmilesTraversal

def _traverse(atom, traverse, prevAtom,
              visitedAtoms, visitedBonds,
              atoms, bonds, Traversal, bondIndex=0):
    visitedAtoms[atom] = 1
    traverse.addAtom(atom)
    atoms.append(atom)

    bondsToTraverse = []
    traversals = []
    
    for bond in atom.bonds:
        oatom = bond.xatom(atom)
        if prevAtom is not None and oatom == prevAtom:
            # we are traversing back the way we came!
            # so don't...
            pass
        elif oatom in visitedAtoms:
            # a closure!
            traverse.addClosure(atom, oatom, bond)
            bonds.append(bond)
            visitedBonds[bond] = 1
        else:
            bondsToTraverse.append((oatom.symorder,
                                    bond.equiv_class,
                                    bondIndex,
                                    oatom,
                                    bond))
        bondIndex += 1

    if not bondsToTraverse:
        # dead end, return
        return

    bondsToTraverse.sort()

    for symorder, bondEclass, index, oatom, obond in bondsToTraverse:
        if oatom in visitedAtoms:
            # somehow, we've seen this atom so skip it
            continue
        nextTraverse = Traversal(traverse)
        traversals.append(nextTraverse)
        nextTraverse.addBond(obond)
        bonds.append(obond)
        visitedBonds[obond] = 1

        _traverse(oatom, nextTraverse, atom,
                  visitedAtoms, visitedBonds,
                  atoms, bonds, Traversal, bondIndex)

    for t in traversals[:-1]:
        traverse.addBranch()
        traverse.append(t)
        traverse.addBranchEnd()
        
    for t in traversals[-1:]:
        traverse.append(t)

def _get_lowest_symorder(atoms):
    best = atoms[0]
    for atom in atoms[1:]:
        if atom.symorder < best.symorder:
            best = atom
    return best

def draw(molecule, TraversalType=SmilesTraversal):
    """(molecule)->canonical representation of a molecule
    Well, it's only canonical if the atom symorders are
    canonical, otherwise it's arbitrary.

    atoms must have a symorder attribute
    bonds must have a equiv_class attribute"""
    result = []
    atoms = allAtoms = molecule.atoms

    visitedAtoms = {}
    #
    # Traverse all components of the graph to form
    # the output string
    while atoms:
        atom = _get_lowest_symorder(atoms)
        visitedAtoms[atom] = 1

        visitedBonds = {}
        nextTraverse = TraversalType()
        atomsUsed, bondsUsed = [], []
        _traverse(atom, nextTraverse, None,
                  visitedAtoms, visitedBonds,
                  atomsUsed, bondsUsed, TraversalType)
        atoms = []
        for atom in allAtoms:
            if not atom in visitedAtoms:
                atoms.append(atom)
        assert nextTraverse.atoms == atomsUsed
        assert nextTraverse.bonds == bondsUsed, "%s %s"%(
            nextTraverse.bonds, bondsUsed)
        

        result.append((str(nextTraverse),
                       atomsUsed, bondsUsed))

    result.sort()
    fragments = []
    for r in result:
        fragments.append(r[0])

    return ".".join(fragments), result

def drawSmarts(molecule):
    return draw(molecule, TraversalType=SmartsTraversal)

def drawIsomeric(molecule):
    return draw(molecule, TraversalType=IsomericSmilesTraversal)



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Default element properties:

The information in an element is:
  number == atomic number
  symbol == the short symbol (eg, He for helium)
  name == the full (American) English name
  mass == the atomic mass, in amu, natural propensities
  negativity == Pauling negativity value
  valences == list of possible valence occupancies

This was pretty much lifted from element.py
Thanks Andrew!
"""
# XXX FIX ME
# Let's make dictionaries instead.  That will
#  be cleaner
class AtomType:
    """AtomType
    This class essentiall reads a bunch of named arguments in the form
    a=2, b=3, c=[]
    and makes a class O with attributes O.a=2, O.b=3 O.c=[]
    Python is cool."""
    def __init__(self, symbol, **args):
        for kw, val in args.items():
            self.__dict__[kw] = val
        self.symbol = symbol
    def __str__(self):
        return "%s"%self.symbol
    
defaultAtomTypes = {}

# these represent valences in various rows of the periodic table.
valence_unknown = None
valence_1 =(1,)
valence_B =(3,)
valence_C =(4,)
# Nitrogen should not be pentavalent
valence_N =(3, 5)
valence_O =(2,)
valence_P =(3, 5)
valence_S =(2, 4, 6)

num = 0
# the big list of properties
for element in (
  ("*" , 0, "unknown", 0.0),
  ("H" , 1.00794, "hydrogen", 2.20),          # 1
  ("He", 4.003, "helium", 0.0),
  ("Li", 6.941, "lithium", 0.98),
  ("Be", 9.0122, "beryllium", 1.57),
  ("B" , 10.81, "boron", 2.04),
  ("C" , 12.011, "carbon", 2.55),
  ("N" , 14.007, "nitrogen", 3.04),
  ("O" , 15.999, "oxygen", 3.44),
  ("F" , 18.998, "fluorine", 3.98),
  ("Ne", 20.179, "neon", 0.0),
  ("Na", 22.990, "sodium", 0.93),            # 11
  ("Mg", 24.305, "magnesium", 1.31),
  ("Al", 26.98, "aluminum", 1.61),
  ("Si", 28.086, "silicon", 1.90),
  ("P" , 30.974, "phosphorus", 2.19),
  ("S" , 32.066, "sulfer", 2.58),
  ("Cl", 35.453, "chlorine", 3.16),
  ("Ar", 39.948, "argon", 0.0),
  ("K" , 39.098, "potassium", 0.82),
  ("Ca", 40.08, "calcium", 1.00),
  ("Sc", 44.956, "scandium", 1.36),           # 21
  ("Ti", 47.88, "titanium", 1.54),
  ("V" , 50.94, "vanadium", 1.63),
  ("Cr", 51.996, "chromium", 1.66),
  ("Mn", 54.938, "manganese", 1.55),
  ("Fe", 55.847, "iron", 1.83),
  ("Co", 58.9332, "cobalt", 1.88),
  ("Ni", 58.69, "nickel", 1.91),
  ("Cu", 63.546, "copper", 1.90),
  ("Zn", 65.39, "zinc", 1.65),
  ("Ga", 69.72, "gallium", 1.81),             # 31
  ("Ge", 72.59, "germanium", 2.01),
  ("As", 74.922, "arsenic", 2.18),
  ("Se", 78.96, "selenium", 2.55),
  ("Br", 79.904, "bromine", 2.96),
  ("Kr", 83.80, "krypton", 0.0),
  ("Rb", 85.468, "rubidium", 0.82),
  ("Sr", 87.62, "strontium", 0.95),
  ("Y" , 88.9059, "yttrium", 1.22),
  ("Zr", 91.224, "zirconium", 1.33),
  ("Nb", 92.91, "niobium", 1.6),              # 41
  ("Mo", 95.94, "molybdenum", 2.16),
  ("Tc", 98., "technetium", 1.9),
  ("Ru", 101.07, "ruthenium", 2.2),
  ("Rh", 102.906, "rhodium", 2.28),
  ("Pd", 106.42, "palladium", 2.20),
  ("Ag", 107.868, "silver", 1.93),
  ("Cd", 112.41, "cadmium", 1.69),
  ("In", 114.82, "indium", 1.78),
  ("Sn", 118.71, "tin", 1.96),
  ("Sb", 121.76, "antimony", 2.05),           # 51
  ("Te", 127.60, "tellurium", 2.1),
  ("I" , 126.9045, "iodine", 2.66),
  ("Xe", 131.29, "xenon", 0.0),
  ("Cs", 132.91, "cesium", 0.79),
  ("Ba", 137.33, "barium", 0.89),
  ("La", 138.906, "lanthanum", 1.10),
  ("Ce", 140.12, "cerium", 1.12),
  ("Pr", 140.908, "praseodymium", 1.13),
  ("Nd", 144.24, "neodymium", 1.14),
  ("Pm", 145., "promethium", 0.0),            # 61
  ("Sm", 150.36, "samarium", 1.17),
  ("Eu", 151.96, "europium", 0.0),
  ("Gd", 157.25, "gadolinium", 1.20),
  ("Tb", 158.925, "terbium", 0.0),
  ("Dy", 162.50, "dysprosium", 1.22),
  ("Ho", 164.93, "holmium", 1.23),
  ("Er", 167.26, "erbium", 1.24),
  ("Tm", 168.934, "thulium", 1.25),
  ("Yb", 173.04, "ytterbium", 0.0),
  ("Lu", 174.967, "lutetium", 1.27),          # 71
  ("Hf", 178.49, "hafnium", 1.3),
  ("Ta", 180.95, "tantalum", 1.5),
  ("W" , 183.84, "tungsten", 2.36),
  ("Re", 186.207, "rhenium", 1.9),
  ("Os", 190.2, "osmium", 2.2),
  ("Ir", 192.22, "iridium", 2.20),
  ("Pt", 195.08, "platinum", 2.28),
  ("Au", 196.967, "gold", 2.54),
  ("Hg", 200.59, "mercury", 2.00),
  ("Tl", 204.383, "thallium", 1.62),            # 81
  ("Pb", 207.2, "lead", 1.8),
  ("Bi", 208.98, "bismuth", 2.02),
  ("Po", 209., "polonium", 2.0),
  ("At", 210., "astatine", 2.2),
  ("Rn", 222., "radon", 0.0),
  ("Fr", 223., "francium", 0.7),
  ("Ra", 226.025, "radium", 0.9),
  ("Ac", 227.028, "actinium", 1.1),
  ("Th", 232.038, "thorium", 1.3),
  ("Pa", 231.036, "protactinium", 1.5),          # 91
  ("U" , 238.029, "uranium", 1.38),
  ("Np", 237.048, "neptunium", 1.36),
  ("Pu", 244., "plutonium", 1.28),
  ("Am", 243., "americium", 1.3),
  ("Cm", 247., "curium", 1.3),
  ("Bk", 247., "berkelium", 1.3),
  ("Cf", 251., "califorium", 1.3),
  ("Es", 252., "einsteinium", 1.3),
  ("Fm", 257., "fermium", 1.3),
  ("Md", 258., "mendelevium", 1.3),           # 101
  ("No", 259., "nobelium", 1.3),
  ("Lr", 260., "lawrencium", 0.0), 
  ("Rf", 261., "rutherfordium", 0.0),
  ("Ha", 262., "hahnium", 0.0),     # also called "dubnium"
  ("Sg", 263., "seagorbium", 0.0),  # once 'unnilhexium'
  ("Ns", 269., "bohrium", 0.0),     # or "nielsbohrium"
  ("Hs", 268., "hassium", 0.0),     #  so what names do you want?
  ("Mt", 266., "meitnerium", 0.0),
  ("Uun", 269., "ununnilium", 0.0),
  ("Uuu", 272., "unununium", 0.0),             #  111
  ("Uub", 277., "ununbium", 0.0),
  # ("Uut", 0.0, "ununtrium", 0.0),   # enter when they are
  # ("Uuq", 0.0, "ununquadium", 0.0), # discovered
  # ("Uup", 0.0, "", 0.0),
  # ("Uuh", 0.0, "", 0.0),
  # ("Uus", 0.0, "", 0.0),
  # ("Uuo", 0.0, "", 0.0),
  ("R", 0, "R Group", 0.0),
  ):
    if num in (1, 9, 17, 35, 53):
        valences = valence_1
    elif num == 5:
        valences = valence_B
    elif num == 6:
        valences = valence_C
    elif num == 7:
        valences = valence_N
    elif num == 8:
        valences = valence_O
    elif num == 15:
        valences = valence_P
    elif num == 16:
        valences = valence_S
    else:
        valences = valence_unknown

    # ok, we have the info, now create the class and
    #  add it to the table
    ele = (element[0],
           num,          # number
           element[2],   # fullname
           element[1],   # mass
           element[3],   # negativity
           valences,     # valences
           num,          # equiv_class
           )
    defaultAtomTypes[element[0]] = ele
    num = num + 1


del valence_unknown
del valence_1
del valence_B
del valence_C
del valence_N
del valence_O
del valence_P
del valence_S
del num
del ele
del valences



"""
FROWNS LICENSE

Copyright (c) 2001-2003, Brian Kelley
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met: 

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer. 
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials  provided with the
      distribution. 
    * Neither the name of Brian Kelley nor the names of frowns
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Ring Perception Algorithm ala figueras

This code has been tested against a slower version of ring detection
similar to the ones that the Babel converter uses.

This code is MUCH faster although their appears to be deficencies
in this ring detection code that need investigating.

On the plus side, it generates the ringsets that can be used for
aromaticity detection very quickly (tested on 300k NCI compounds,
the canonicalization using aromaticity detection was the same for
both ring detectors).

On the minus side, there may be some bugs in the code that may limit
its use as a full smallest set of smallest rings finder.

This is also the ring detection code that CDK uses for their ring
perception so that might be a branch to investigate.
"""
from .cycle import Cycle
from itertools import zip_longest
#from CheckFiguerasRings import checkRings

def sssr(molecule):
    """molecule -> generate the molecule.cycles that contain
    the smallest set of smallest rings"""
    results = {}
    lookup = {}
    fullSet = {}
    oatoms = {}
    # XXX FIX ME
    # copy atom.oatoms to atom._oatoms
    # atom._oatoms will be modified my the routine

    for atom in molecule.atoms:
        atom.rings = []
        fullSet[atom.handle] = 1
        lookup[atom.handle] = atom
        oatoms[atom.handle] = atom.oatoms[:]

    for bond in molecule.bonds:
        bond.rings = []
        
    trimSet = []

    while fullSet:
        nodesN2 = []
        minimum, minimum_degree = None, 100000


        # find the N2 atoms and remove atoms with degree 0
        for atomID in list(fullSet):
            atom = lookup[atomID]
            degree = len(oatoms[atom.handle])
            if degree == 0:
                del fullSet[atomID]
                #fullSet.remove(atomID)
            elif degree == 2:
                nodesN2.append(atom)

            # keep track of the minimum degree
            if (degree > 0) and ( (not minimum) or 
                                (degree < minimum_degree)):
                minimum, minimum_degree = atom, degree

        if not minimum:
            # nothing to do!  (i.e. can't have a ring)
            break

        if minimum_degree == 1:
            # these cannot be in rings so trim and remove
            # my version of trimming
            for oatom in oatoms[minimum.handle]:
                oatoms[oatom.handle].remove(minimum)
            oatoms[minimum.handle] = []
            del fullSet[minimum.handle]

        elif minimum_degree == 2:
            # find the rings!
            startNodes = []
            for atom in nodesN2:
                ring, bonds = getRing(atom, fullSet, lookup, oatoms)
                ring = list(ring)

                if ring:
                    rlookup = ring[:]
                    rlookup.sort()
                    rlookup = tuple(rlookup)
                    if (not rlookup in results):# not in results):
                        results[rlookup] = ring, bonds
                        startNodes.append(atom)

            # in case we didn't get a ring remove the head of the nodesN2
            startNodes = startNodes or [nodesN2[0]]
            for atom in startNodes:
                # again, my version of trimming
                if oatoms[atom.handle]:
                    oatom = oatoms[atom.handle].pop()
                    oatoms[oatom.handle].remove(atom)

        elif minimum_degree > 2:
            # no N2 nodes so remove the "optimum" edge to create
            # N2 nodes in the next go-around.
            ring, bonds = getRing(minimum, fullSet, lookup, oatoms)
            ring = list(ring)
            if ring:
                key = ring[:]
                key.sort()
                key = tuple(key)
                if not key in results:
                    results[key] = ring, bonds
                    atoms = map(lookup.get, ring)
                    atoms, bonds = toposort(atoms, bonds)
                    checkEdges(atoms, lookup, oatoms)
            else:
                del fullSet[minimum.handle]
        else:
            raise ShouldntGetHereError

    # assign the ring index to the atom
    rings = []
    index = 0

    # transform the handles back to atoms
    for result, bonds in results.values():
        ring = []
        for atomID in result:
            atom = lookup[atomID]
            assert atom.handle == atomID
            ring.append(atom)
        rings.append((ring, bonds))
        index = index + 1

    molecule.rings = rings
    potentialCycles = []
    index = 0
    for atoms, bonds in rings:
        # due to the dictionaries used in getRing
        # the atoms are not in the order found
        # we need to topologically sort these
        # for the cycle
        atoms, bonds = toposort(atoms, bonds)
        potentialCycles.append((atoms, bonds))

    rings = potentialCycles#checkRings(potentialCycles)
    molecule.rings = rings
    molecule.cycles = [Cycle(atoms, bonds) for atoms, bonds in rings]
    return molecule

def toposort(initialAtoms, initialBonds):
    """initialAtoms, initialBonds -> atoms, bonds
    Given the list of atoms and bonds in a ring
    return the topologically sorted atoms and bonds.
    That is each atom is connected to the following atom
    and each bond is connected to the following bond in
    the following manner
    a1 - b1 - a2 - b2 - ... """
    atoms = []
    a_append = atoms.append
    bonds = []
    b_append = bonds.append

    # for the atom and bond hashes
    # we ignore the first atom since we
    # would have deleted it from the hash anyway
    ahash = {}
    bhash = {}
    initialAtoms = list(initialAtoms)
    for atom in initialAtoms[1:]:
        ahash[atom.handle] = 1
        
    for bond in initialBonds:
        bhash[bond.handle] = bond

    next = initialAtoms[0]
    a_append(next)

    # do until all the atoms are gone
    while ahash:
        # traverse to all the connected atoms
        for atom in next.oatoms:
            # both the bond and the atom have to be
            # in our list of atoms and bonds to use
            # ugg, nested if's...  There has to be a
            # better control structure
            if atom.handle in ahash:
                bond = next.findbond(atom)
                assert bond
                # but wait! the bond has to be in our
                # list of bonds we can use!
                if bond.handle in bhash:
                    a_append(atom)
                    b_append(bond)
                    del ahash[atom.handle]
                    next = atom
                    break
        else:
            raise RingException("Atoms are not in ring")

    assert len(initialAtoms) == len(atoms)
    assert len(bonds) == len(atoms) - 1
    lastBond = atoms[0].findbond(atoms[-1])
    assert lastBond
    b_append(lastBond)
    return atoms, bonds
    
def getRing(startAtom, atomSet, lookup, oatoms):
    """getRing(startAtom, atomSet, lookup, oatoms)->atoms, bonds
    starting at startAtom do a bfs traversal through the atoms
    in atomSet and return the smallest ring found

    returns (), () on failure
    note: atoms and bonds are not returned in traversal order"""

    path = {}
    bpaths = {}
    for atomID in atomSet.keys():
        # initially the paths are empty
        path[atomID] = None
        bpaths[atomID] = []
    
    q = []
    handle = startAtom.handle
    for atom in oatoms[handle]:
        q.append((atom, handle))
        path[atom.handle] = {atom.handle:1, handle:1}
        bpaths[atom.handle] = [startAtom.findbond(atom)]
            
    qIndex = 0
    lenQ = len(q)

    while qIndex < lenQ:
        current, sourceHandle = q[qIndex]
        handle = current.handle
        qIndex += 1

        for next in oatoms[handle]:
            m = next.handle

            if m != sourceHandle:
                if not m in atomSet:
                    return (), ()
                
                if path.get(m, None):
                    intersections = 0
                    for atom in path[handle].keys():
                        if atom in path[m]:
                            intersections = intersections + 1
                            sharedAtom = atom

                    if intersections == 1:
                        del path[handle][sharedAtom]
                        path[handle].update(path[m])
                        result = path[handle].keys()
                        bond = next.findbond(current)
                        # assert bond not in bpaths[handle] and bond not in bpaths[m]
                        bonds = bpaths[handle] + bpaths[m] + [bond]
                        return result, bonds
                else:
                    path[m] = path[handle].copy()
                    path[m][m] = 1
                    bond = next.findbond(current)
                    # assert bond not in bpaths[m] and bond not in bpaths[handle]
                    bpaths[m] = bpaths[handle] + [next.findbond(current)]
                    q.append((next, handle))
                    lenQ = lenQ + 1

    return (), ()


def checkEdges(ringSet, lookup, oatoms):
    """atoms, lookup -> ring
    atoms must be in the order of traversal around a ring!
    break an optimal non N2 node and return the largest ring
    found
    """
    bondedAtoms = list(zip_longest(ringSet[:-1], ringSet[1:]))
    bondedAtoms += [ (ringSet[-1], ringSet[0]) ]

    # form a lookup for the ringSet list
    atomSet = {}
    for atomID in ringSet:
        atomSet[atomID] = 1
    results = []
    
    # for each bond in the ring, break it and find the smallest
    # rings starting on either side of the bond
    # keep the largest but rememeber to add the bond back at the
    # end
    for atom1, atom2 in bondedAtoms:
        # break a single edge in the ring
        handle1 = atom1.handle
        handle2 = atom2.handle
        oatoms1 = oatoms[handle1]
        oatoms2 = oatoms[handle2]
        index1 = oatoms1.index(atom2)
        index2 = oatoms2.index(atom1)

        # break the bond
        del oatoms1[index1]
        del oatoms2[index2]

        ring1 = getRing(atom1, atomSet, lookup, oatoms)
        ring2 = getRing(atom2, atomSet, lookup, oatoms)
        
        # keep the larger of the two rings
        if len(ring1) > len(ring2):
            results.append((len(ring1),
                            handle1, handle2,
                            ring1))
        else:
            results.append((len(ring2),
                            handle2, handle1,
                            ring2))

        # retie the bond
        oatoms1.insert(index1, atom2)
        oatoms2.insert(index2, atom1)


    if not results:
        return None

    #  find the smallest ring
    size, incidentHandle, adjacentHandle, smallestRing = min(results)
    # dereference the handles
    incident, adjacent = lookup[incidentHandle], lookup[adjacentHandle]

    # break the bond between the incident and adjacent atoms
    oatomsI = oatoms[incidentHandle]
    oatomsA = oatoms[adjacentHandle]
    assert incident in oatomsA
    assert adjacent in oatomsI
    
    oatomsI.remove(adjacent)
    oatomsA.remove(incident)



from setuptools import setup, find_packages

VERSION = '0.0.3'

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pinky',
    description='pinky - molecular fingerprint library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andrew E. Bruno',
    url='https://github.com/ubccr/pinky',
    license='BSD',
    author_email='aebruno2@buffalo.edu',
    include_package_data=True,
    version=VERSION,
    install_requires=[
        'bitarray'
    ],
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
     ]
)
